clear; close all; clc;
import casadi.*;


% parameters
N  = 40;
m = 4/N;
D = (70/40)*N;
g = 9.81;

opti = casadi.Opti();
y = opti.variable(N);
z = opti.variable(N);

% objective
obj = .5 * D * sum(diff(y).^2 + diff(z).^2);
obj = obj + g * m * sum(z);
opti.minimize(obj);

% end point constraints
opti.subject_to(y(1) == -2);
opti.subject_to(z(1) == 1);
opti.subject_to(y(end) == 2);
opti.subject_to(z(end) == 1);

% floor constraint
opti.subject_to(z + y.^2 >= 0);

opti.callback( @(i) plot_chain(full(opti.value(y)),full(opti.value(z)), i) )

opti.solver('ipopt');
sol = opti.solve();


%% LICQ and active set

% active inequalities
g_sol = sol.value(opti.g);
act_ineq = find(abs(g_sol) <= 1e-6) ;

% add equalities
act = [(1:4)'; act_ineq];

% jacobian at sol
g_jac = jacobian(opti.g, opti.x);
g_jac_sol = sol.value(g_jac);

% jacobian of active constraints
g_jac_act = full(g_jac_sol(act, :));
fprintf('\n\n')
fprintf('size of active jacobian: %dx%d\n', size(g_jac_act,1), size(g_jac_act,2))
fprintf('rank of active jacobian: %d\n', rank(g_jac_act))

% lagrange multipliers
lam_g = sol.value(opti.lam_g);

% multipliers of active inequalities
mu_act = lam_g(act_ineq);
% check which are weakly active
mu_weak = mu_act(abs(mu_act) <= 1e-7);

fprintf('number of weakly active constraints: %d\n', length(mu_weak))

%% Hessian of Lagrangian

% lagrangian
L = opti.f + opti.lam_g' * opti.g;

% hessian
L_hess = hessian(L, opti.x);
L_hess_sol = sol.value(L_hess);


%% obtain null space

% using QR
[Q,R] = qr(g_jac_act');
Z = Q(:,length(act)+1:end);

% using SVD
% [U,S,V] = svd(g_jac_act);
% Z = V(:,length(act)+1:end);

% check nullspace
res = max(max(abs(g_jac_act * Z)));       % should be approx. zero

% alternatively:
% Z = null(g_jac_act);

%% reduced Hessian
redH = Z' * L_hess_sol * Z;
fprintf('smallest EV of reduced Hessian: %f\n', min(eig(redH)));

